@@ -44,7 +44,7 @@ class OpenADRClient:
     """
     def __init__(self, ven_name, vtn_url, debug=False, cert=None, key=None,
                  passphrase=None, vtn_fingerprint=None, show_fingerprint=True, ca_file=None,
-                 allow_jitter=True, ven_id=None, disable_signature=False):
+                 allow_jitter=True, ven_id=None, disable_signature=False, check_hostname=True):
         """
         Initializes a new OpenADR Client (Virtual End Node)
 
@@ -68,14 +68,13 @@ class OpenADRClient:
         """
 
         self.ven_name = ven_name
-        if vtn_url.endswith("/"):
-            vtn_url = vtn_url[:-1]
-        self.vtn_url = vtn_url
+        self.vtn_url = vtn_url.rstrip("/")
         self.ven_id = ven_id
         self.registration_id = None
         self.poll_frequency = None
         self.vtn_fingerprint = vtn_fingerprint
         self.debug = debug
+        self.check_hostname = check_hostname
 
         self.reports = []
         self.report_callbacks = {}              # Holds the callbacks for each specific report
@@ -115,6 +114,11 @@ class OpenADRClient:
                                        key=key,
                                        passphrase=passphrase,
                                        disable_signature=disable_signature)
+        self.hooks = {'before_send_xml': [],
+                      'after_receive_xml': [],
+                      'before_schema_validation': [],
+                      'before_parse_xml': [],
+                      'after_parse_xml': []}
 
     async def run(self):
         """
@@ -130,10 +134,14 @@ class OpenADRClient:
             await self.stop()
             return
 
+        await self.register_reports(self.reports)
         if self.reports:
-            await self.register_reports(self.reports)
             self.report_queue_task = self.loop.create_task(self._report_queue_worker())
 
+        # Perform initial event sync
+        await self.sync_events()
+
+        # Perform an initial poll
         await self._poll()
 
         # Set up automatic polling
@@ -141,11 +149,10 @@ class OpenADRClient:
             logger.warning("Polling with intervals of more than 24 hours is not supported. "
                            "Will use 24 hours as the polling interval.")
             self.poll_frequency = timedelta(hours=24)
-        cron_config = utils.cron_config(self.poll_frequency, randomize_seconds=self.allow_jitter)
 
         self.scheduler.add_job(self._poll,
-                               trigger='cron',
-                               **cron_config)
+                               trigger='interval',
+                               seconds=self.poll_frequency.total_seconds())
         self.scheduler.add_job(self._event_cleanup,
                                trigger='interval',
                                seconds=300)
@@ -324,7 +331,7 @@ class OpenADRClient:
 
         # Add the new report description to the report
         target = objects.Target(resource_id=resource_id)
-        r_id = utils.generate_id()
+        r_id = r_id or utils.generate_id()
         report_description = objects.ReportDescription(r_id=r_id,
                                                        reading_type=reading_type,
                                                        report_data_source=target,
@@ -337,6 +344,21 @@ class OpenADRClient:
         report.report_descriptions.append(report_description)
         return report_specifier_id, r_id
 
+    def add_hook(self, hook_name, handler):
+        """
+        You can add a hook at specific points in the request/reponse chain.
+        Your choices are:
+
+        before_send_xml: you get the actual XML message just before it is sent over the wire
+        after_receive_xml: you get the actual XML immediately after it is received over the wire
+        before_parse_xml: you get the actual XML after its schema has been validated, but before parsing
+        after_parse_xml: you get the message name and message payload after parsing the message
+        """
+        if hook_name not in self.hooks:
+            raise ValueError(f"The hook_name should be one of {', '.join(self.hooks.keys())}. "
+                             f"You provided: {hook_name}.")
+        self.hooks[hook_name].append(handler)
+
     ###########################################################################
     #                                                                         #
     #                             POLLING METHODS                             #
@@ -472,6 +494,14 @@ class OpenADRClient:
         message = self._create_message('oadrCreatedEvent', **payload)
         response_type, response_payload = await self._perform_request(service, message)
 
+    async def sync_events(self):
+        """
+        Used to perform an initial sync of events after the client connects
+        """
+        response_type, response_payload = await self.request_event()
+        if 'events' in response_payload and len(response_payload['events']) > 0:
+            await self._on_event(response_payload)
+
     ###########################################################################
     #                                                                         #
     #                             REPORTING METHODS                           #
@@ -501,17 +531,18 @@ class OpenADRClient:
             for report_request in response_payload['report_requests']:
                 await self.create_report(report_request)
 
-        # Send the oadrCreatedReport message
-        message_type = 'oadrCreatedReport'
-        message_payload = {'pending_reports':
-                           [{'report_request_id': utils.getmember(report, 'report_request_id')}
-                            for report in self.report_requests]}
-        message = self._create_message(message_type,
-                                       response={'response_code': 200,
-                                                 'response_description': 'OK'},
-                                       ven_id=self.ven_id,
-                                       **message_payload)
-        response_type, response_payload = await self._perform_request(service, message)
+            # Send the oadrCreatedReport message
+            message_type = 'oadrCreatedReport'
+            message_payload = {'pending_reports':
+                               [{'report_request_id': utils.getmember(report, 'report_request_id')}
+                                for report in self.report_requests]}
+            message = self._create_message(message_type,
+                                           response={'response_code': 200,
+                                                     'response_description': 'OK',
+                                                     'request_id': response_payload['response']['request_id']},
+                                           ven_id=self.ven_id,
+                                           **message_payload)
+            response_type, response_payload = await self._perform_request(service, message)
 
     async def create_report(self, report_request):
         """
@@ -721,12 +752,7 @@ class OpenADRClient:
                        "an Event dict and should return either 'optIn' or 'optOut' based on your "
                        "choice. Will re-use the previous opt status for this event_id for now")
         if event['event_descriptor']['event_id'] in self.responded_events:
-            return self.responded_events['event_id']
-
-    async def on_register_report(self, report):
-        """
-        Placeholder for the on_register_report handler.
-        """
+            return self.responded_events.get(event['event_descriptor']['event_id'])
 
     async def on_cancel_party_registration(self, message):
         # Update/Delete all the registration and reports information
@@ -773,8 +799,10 @@ class OpenADRClient:
         await self._ensure_client_session()
         url = f"{self.vtn_url}/{service}"
         try:
+            await self._execute_hooks('before_send_xml', utils.ensure_str(message))
             async with self.client_session.post(url, data=message) as req:
                 content = await req.read()
+                await self._execute_hooks('after_receive_xml', utils.ensure_str(content))
                 if req.status != HTTPStatus.OK:
                     logger.warning(f"Non-OK status {req.status} when performing a request to {url} "
                                    f"with data {message}: {req.status} {content.decode('utf-8')}")
@@ -790,10 +818,13 @@ class OpenADRClient:
         if len(content) == 0:
             return None
         try:
+            await self._execute_hooks('before_schema_validation', utils.ensure_str(content))
             tree = validate_xml_schema(content)
             if self.vtn_fingerprint:
                 validate_xml_signature(tree, cert_fingerprint=self.vtn_fingerprint)
+            await self._execute_hooks('before_parse_xml', utils.ensure_str(content))
             message_type, message_payload = parse_message(content)
+            await self._execute_hooks('after_parse_xml', message_type, message_payload)
         except XMLSyntaxError as err:
             logger.warning(f"Incoming message did not pass XML schema validation: {err}")
             return None, {}
@@ -813,6 +844,14 @@ class OpenADRClient:
                                f"{message_payload['response']['response_description']}")
         return message_type, message_payload
 
+    async def _execute_hooks(self, hook_name, *args, **kwargs):
+        for hook in self.hooks[hook_name]:
+            try:
+                await utils.await_if_required(hook(*args, **kwargs))
+            except Exception as err:
+                logger.error(f"An error occurred while executing your '{hook_name}': {hook}:"
+                             f"{err.__class__.__name__}: {err}")
+
     async def _on_event(self, message):
         logger.debug("The VEN received an event")
         events = message['events']
@@ -907,7 +946,6 @@ class OpenADRClient:
 
         elif response_type == 'oadrDistributeEvent':
             if 'events' in response_payload and len(response_payload['events']) > 0:
-                logger.info(f"The payload tyupe {type(response_payload)}")
                 await self._on_event(response_payload)
 
         elif response_type == 'oadrUpdateReport':
@@ -919,9 +957,16 @@ class OpenADRClient:
                     await self.create_report(report_request)
 
         elif response_type == 'oadrRegisterReport':
-            if 'reports' in response_payload and len(response_payload['reports']) > 0:
-                for report in response_payload['reports']:
-                    await self.register_report(report)
+            # We don't support receiving reports from the VTN at this moment
+            logger.warning("The VTN offered reports, but OpenLEADR "
+                           "does not support reports in this direction.")
+            message = self._create_message('oadrRegisteredReport',
+                                           report_requests=[],
+                                           response={'response_code': 200,
+                                                     'response_description': 'OK',
+                                                     'request_id': response_payload['request_id']})
+            service = 'EiReport'
+            reponse_type, response_payload = await self._perform_request(service, message)
 
         elif response_type == 'oadrCancelPartyRegistration':
             logger.info("The VTN required us to cancel the registration. Calling the cancel party registration procedure.")
@@ -937,12 +982,20 @@ class OpenADRClient:
     async def _ensure_client_session(self):
         if not self.client_session:
             headers = {'content-type': 'application/xml'}
+            client_timeout = aiohttp.ClientTimeout(sock_connect=5, sock_read=10)
             if self.cert_path:
-                ssl_context = ssl.create_default_context(cafile=self.ca_file,
-                                                         purpose=ssl.Purpose.CLIENT_AUTH)
+                ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
+                ssl_context.load_verify_locations(self.ca_file)
                 ssl_context.load_cert_chain(self.cert_path, self.key_path, self.passphrase)
-                ssl_context.check_hostname = False
+                ssl_context.check_hostname = self.check_hostname
                 connector = aiohttp.TCPConnector(ssl=ssl_context)
-                self.client_session = aiohttp.ClientSession(connector=connector, headers=headers)
+                self.client_session = aiohttp.ClientSession(
+                    connector=connector,
+                    headers=headers,
+                    timeout=client_timeout
+                )
             else:
-                self.client_session = aiohttp.ClientSession(headers=headers)
+                self.client_session = aiohttp.ClientSession(
+                    headers=headers,
+                    timeout=client_timeout
+                )
