@@ -22,7 +22,6 @@ from uuid import uuid4
 from lxml.etree import Element
 from openleadr import errors
 from datetime import datetime, timezone, timedelta
-from openleadr.utils import ensure_bytes
 import os
 
 from openleadr import utils
@@ -50,13 +49,20 @@ def parse_message(data):
 
     Returns a message type (str) and a message payload (dict)
     """
+    try:
+        if isinstance(data, bytes):
+            logger.debug(f"Parsing message: {data.decode('utf-8')}")
+        else:
+            logger.debug(f"Parsing message: {data}")
+    except UnicodeDecodeError:
+        logger.warning(f"Could not decode incoming message as UTF-8: {str(data)}")
     message_dict = xmltodict.parse(data, process_namespaces=True, namespaces=NAMESPACES)
     message_type, message_payload = message_dict['oadrPayload']['oadrSignedObject'].popitem()
     message_payload = utils.normalize_dict(message_payload)
     return message_type, message_payload
 
 
-def create_message(message_type, cert=None, key=None, passphrase=None, **message_payload):
+def create_message(message_type, cert=None, key=None, passphrase=None, disable_signature=False, **message_payload):
     """
     Create and optionally sign an OpenADR message. Returns an XML string.
     """
@@ -64,7 +70,7 @@ def create_message(message_type, cert=None, key=None, passphrase=None, **message
     template = TEMPLATES.get_template(f'{message_type}.xml')
     signed_object = utils.flatten_xml(template.render(**message_payload))
     envelope = TEMPLATES.get_template('oadrPayload.xml')
-    if cert and key:
+    if cert and key and not disable_signature:
         tree = etree.fromstring(signed_object)
         signature_tree = SIGNER.sign(tree,
                                      key=key,
@@ -75,11 +81,10 @@ def create_message(message_type, cert=None, key=None, passphrase=None, **message
         signature = etree.tostring(signature_tree).decode('utf-8')
     else:
         signature = None
-
     msg = envelope.render(template=f'{message_type}',
                           signature=signature,
                           signed_object=signed_object)
-    validate_xml_schema(ensure_bytes(msg))
+    logger.debug(f"Created message: {msg}")
     return msg
 
 
@@ -108,6 +113,10 @@ def validate_xml_signature(xml_tree, cert_fingerprint=None):
     _verify_replay_protect(xml_tree)
 
 
+def validate_xml_signature_none(xml_tree):
+    assert xml_tree.find('.//{http://www.w3.org/2000/09/xmldsig#}X509Certificate') is None
+
+
 async def authenticate_message(request, message_tree, message_payload, fingerprint_lookup=None, ven_lookup=None):
     if request.secure and 'ven_id' in message_payload:
         connection_fingerprint = utils.get_cert_fingerprint_from_request(request)
@@ -143,10 +152,17 @@ async def authenticate_message(request, message_tree, message_payload, fingerpri
                    f"does not match the expected fingerprint '{expected_fingerprint}'")
             raise errors.NotRegisteredOrAuthorizedError(msg)
 
+        message_cert = utils.extract_pem_cert(message_tree)
+        message_fingerprint = utils.certificate_fingerprint(message_cert)
+        if message_fingerprint != expected_fingerprint:
+            msg = (f"The fingerprint of the certificate used to sign the message "
+                   f"{message_fingerprint} did not match the fingerprint that this "
+                   f"VTN has for you {expected_fingerprint}. Make sure you use the correct "
+                   "certificate to sign your messages.")
+            raise errors.NotRegisteredOrAuthorizedError(msg)
+
         try:
-            validate_xml_signature(message_tree, expected_fingerprint)
-        except errors.FingerprintMismatch as err:
-            raise errors.NotRegisteredOrAuthorizedError(err.msg)
+            validate_xml_signature(message_tree)
         except ValueError:
             msg = ("The message signature did not match the message contents. Please make sure "
                    "you are using the correct XMLDSig algorithm and C14n canonicalization.")
